from functools import wraps
from typing import TYPE_CHECKING, Callable, Optional

from botocraft.services.abstract import PrimaryBoto3ModelQuerySet

if TYPE_CHECKING:
    from botocraft.services.inspector2 import (
        CisScanConfiguration,
        DelegatedAdmin,
        DelegatedAdminAccount,
    )


def convert_delegated_admin(
    func: Callable[..., Optional["DelegatedAdmin"]],
) -> Callable[..., Optional["DelegatedAdminAccount"]]:
    """
    Wraps a boto3 method that returns a list of :py:class:`Reservation` objects
    to return a list of :py:class:`Instance` objects instead.
    """

    @wraps(func)
    def wrapper(*args, **kwargs) -> Optional["DelegatedAdminAccount"]:
        from botocraft.services.inspector2 import DelegatedAdminAccount

        self = args[0]
        acct = func(*args, **kwargs)
        if not acct:
            return None
        accts = [
            _acct
            for _acct in DelegatedAdminAccount.objects.using(self.session).list()
            if _acct.accountId == acct.accountId
        ]
        assert len(accts) == 1, "Expected exactly one account"
        _account = DelegatedAdminAccount(
            accountId=acct.accountId,
            status=accts[0].status,
            relationshipStatus=acct.relationshipStatus,
        )
        _account.set_session(self.session)
        return _account

    return wrapper


def list_augment_delegated_admin_accounts(
    func: Callable[..., "PrimaryBoto3ModelQuerySet"],  # noqa: F821],
) -> Callable[..., "PrimaryBoto3ModelQuerySet"]:
    """
    Wraps a boto3 method that returns a list of :py:class:`Reservation` objects
    to return a list of :py:class:`Instance` objects instead.
    """

    @wraps(func)
    def wrapper(*args, **kwargs) -> "PrimaryBoto3ModelQuerySet":
        from botocraft.services.inspector2 import DelegatedAdminAccount

        self = args[0]
        qs = func(*args, **kwargs)
        new_accts = []
        for acct in qs.results:
            _acct = DelegatedAdminAccount.objects.using(self.session).get(
                delegatedAdminAccountId=acct.accountId  # type: ignore[attr-defined]
            )
            # DelegatedAdminAccount is a frozen (readonly) model, so we need to
            # do a weird workaround to set the relationshipStatus attribute
            new_accts.append(
                _acct.model_copy(
                    update={"relationshipStatus": _acct.relationshipStatus}
                )
            )
        return PrimaryBoto3ModelQuerySet(new_accts)

    return wrapper


class CisScanConfigurationManagerMixin:
    """
    Mixin class for CIS scan configuration model.
    """

    def get(
        self,
        cisScanConfigurationArn: str | None = None,  # noqa: N803
        scanName: str | None = None,  # noqa: N803
    ) -> Optional["CisScanConfiguration"]:
        """
        Get a CIS scan configuration by ARN.

        Keyword Args:
            cisScanConfigurationArn: The ARN of the CIS scan configuration.
            scanName: The name of the CIS scan configuration.

        Returns:
            The CIS scan configuration.

        """
        from botocraft.services.inspector2 import (
            CisScanConfiguration,
            CisStringFilter,
            ListCisScanConfigurationsFilterCriteria,
        )

        assert any([cisScanConfigurationArn, scanName]), (
            "Either cisScanConfigurationArn or scanName must be provided"
        )

        if cisScanConfigurationArn:
            filter_criteria = ListCisScanConfigurationsFilterCriteria(
                scanConfigurationArnFilters=[
                    CisStringFilter(value=cisScanConfigurationArn, comparison="EQUALS")
                ]
            )
        else:
            filter_criteria = ListCisScanConfigurationsFilterCriteria(
                scanNameFilters=[CisStringFilter(value=scanName, comparison="EQUALS")]
            )

        configs = CisScanConfiguration.objects.using(self.session).list(  # type: ignore[attr-defined]
            filterCriteria=filter_criteria,
        )
        return configs.scanConfigurations[0] if configs.scanConfigurations else None


class VulnerabilityManagerMixin:
    """
    Mixin class for vulnerability model.
    """

    def get(self, vulnerabilityId: str) -> "PrimaryBoto3ModelQuerySet":  # noqa: N803
        """
        Get a vulnerability by ID or name.

        Keyword Args:
            vulnerabilityId: The ID of the vulnerability.

        Returns:
            The vulnerability.

        """
        from botocraft.services.inspector2 import (
            SearchVulnerabilitiesFilterCriteria,
            Vulnerability,
        )

        response = Vulnerability.objects.using(self.session).get(  # type: ignore[attr-defined]
            filterCriteria=SearchVulnerabilitiesFilterCriteria(
                vulnerabilityIds=[vulnerabilityId]
            )
        )
        return PrimaryBoto3ModelQuerySet(response.vulnerabilities)  # type: ignore[arg-type]



from functools import wraps
from typing import TYPE_CHECKING, Callable, cast

from botocraft.services.abstract import PrimaryBoto3ModelQuerySet

if TYPE_CHECKING:
    from botocraft.services.iam import AttachedPolicy, IAMPolicy

# ------------------------------------------------------------
# Decorators
# ------------------------------------------------------------


def role_inline_policies_only(
    func: Callable[..., list[str]],
) -> Callable[..., "PrimaryBoto3ModelQuerySet"]:
    """
    Wraps :py:meth:`botocraft.services.iam.IAMRoleManager.list_policies` to
    return a :py:class:`PrimaryBoto3ModelQuerySet` of
    :py:class:`botocraft.services.iam.GetRolePolicyResponse` objects instead of
    only a list of names.
    """

    @wraps(func)
    def wrapper(self, *args, **kwargs) -> "PrimaryBoto3ModelQuerySet":
        # We need to import IAMUser here because otherwise we get a circular import
        from botocraft.services.iam import IAMGroup

        groupname = args[0]
        names = func(self, *args, **kwargs)
        policies = [
            IAMGroup.objects.get_policy(GroupName=groupname, PolicyName=name)
            for name in names
        ]
        return PrimaryBoto3ModelQuerySet(policies)

    return wrapper


def role_attached_policies_only(
    func: Callable[..., list["AttachedPolicy"]],
) -> Callable[..., "PrimaryBoto3ModelQuerySet"]:
    """
    Wraps :py:meth:`botocraft.services.iam.IAMRoleManager.list_attached_policies` to
    return a :py:class:`PrimaryBoto3ModelQuerySet` of
    :py:class:`botocraft.services.iam.IAMPolicy` objects instead of only a list
    of names.
    """

    @wraps(func)
    def wrapper(self, *args, **kwargs) -> "PrimaryBoto3ModelQuerySet":
        # We need to import IAMRole here because otherwise we get a circular import
        from botocraft.services.iam import IAMPolicy

        attached_policies = func(self, *args, **kwargs)
        policies = [
            IAMPolicy.objects.get(policy.PolicyArn) for policy in attached_policies
        ]
        return PrimaryBoto3ModelQuerySet(policies)

    return wrapper


def group_inline_policies_only(
    func: Callable[..., list[str]],
) -> Callable[..., "PrimaryBoto3ModelQuerySet"]:
    """
    Wraps :py:meth:`botocraft.services.iam.IAMGroupManager.list_policies` to
    return a :py:class:`PrimaryBoto3ModelQuerySet` of
    :py:class:`botocraft.services.iam.GetGroupPolicyResponse` objects instead of
    only a list of names.
    """

    @wraps(func)
    def wrapper(self, *args, **kwargs) -> "PrimaryBoto3ModelQuerySet":
        # We need to import IAMUser here because otherwise we get a circular import
        from botocraft.services.iam import IAMGroup

        groupname = args[0]
        names = func(self, *args, **kwargs)
        policies = [
            IAMGroup.objects.get_policy(GroupName=groupname, PolicyName=name)
            for name in names
        ]
        return PrimaryBoto3ModelQuerySet(policies)

    return wrapper


def group_attached_policies_only(
    func: Callable[..., list["AttachedPolicy"]],
) -> Callable[..., "PrimaryBoto3ModelQuerySet"]:
    """
    Wraps :py:meth:`botocraft.services.iam.IAMGroupManager.list_attached_policies` to
    return a :py:class:`PrimaryBoto3ModelQuerySet` of
    :py:class:`botocraft.services.iam.IAMPolicy` objects instead of only a list
    of names.
    """

    @wraps(func)
    def wrapper(self, *args, **kwargs) -> "PrimaryBoto3ModelQuerySet":
        # We need to import IAMGroup here because otherwise we get a circular import
        from botocraft.services.iam import IAMPolicy

        attached_policies = func(self, *args, **kwargs)
        policies = [
            IAMPolicy.objects.get(policy.PolicyArn) for policy in attached_policies
        ]
        return PrimaryBoto3ModelQuerySet(policies)

    return wrapper


def user_inline_policies_only(
    func: Callable[..., list[str]],
) -> Callable[..., "PrimaryBoto3ModelQuerySet"]:
    """
    Wraps :py:meth:`botocraft.services.iam.IAMPolicyManager.list_policies` to
    return a :py:class:`PrimaryBoto3ModelQuerySet` of
    :py:class:`botocraft.services.iam.GetUserPolicyResponse` objects instead of
    only a list of names.
    """

    @wraps(func)
    def wrapper(self, *args, **kwargs) -> "PrimaryBoto3ModelQuerySet":
        # We need to import IAMUser here because otherwise we get a circular import
        from botocraft.services.iam import IAMUser

        username = args[0]
        names = func(self, *args, **kwargs)
        policies = [
            IAMUser.objects.get_policy(UserName=username, PolicyName=name)
            for name in names
        ]
        return PrimaryBoto3ModelQuerySet(policies)

    return wrapper


def iam_attached_policies_only(
    func: Callable[..., list["AttachedPolicy"]],
) -> Callable[..., "PrimaryBoto3ModelQuerySet"]:
    """
    Wraps
    :py:meth:`botocraft.services.iam.IAMPolicyManager.list_attached_policies`
    and any other method that returns a list of IAMPolicy (as opposed to inline
    policy) names to return a :py:class:`PrimaryBoto3ModelQuerySet` of
    :py:class:`botocraft.services.iam.IAMPolicy` objects instead of only a list
    of names.
    """

    @wraps(func)
    def wrapper(self, *args, **kwargs) -> "PrimaryBoto3ModelQuerySet":
        # We need to import IAMPolicy here because otherwise we get a circular import
        from botocraft.services.iam import IAMPolicy

        attached_policies = func(self, *args, **kwargs)
        policies = [
            IAMPolicy.objects.get(policy.PolicyArn) for policy in attached_policies
        ]
        return PrimaryBoto3ModelQuerySet(policies)

    return wrapper


# ------------------------------------------------------------
# Model Mixins
# ------------------------------------------------------------


class IAMPolicyMixin:
    """
    A mixin that adds a ``policy`` property to a model instance.  This
    is intended to be used with :py:class:`Boto3Model` classes that have
    a ``Policy`` attribute.
    """

    @property
    def is_in_use(self) -> bool:
        """
        Return True if the policy is used by any users, roles, or groups.
        """
        entities = self.entities()  # type: ignore[attr-defined]
        return all(
            [
                len(entities.PolicyUsers) > 0,
                len(entities.PolicyRoles) > 0,
                len(entities.PolicyGroups) > 0,
            ]
        )


# ------------------------------------------------------------
# Manager Mixins
# ------------------------------------------------------------


class IAMPolicyManagerMixin:
    """
    A mixin that adds a ``policy`` property to a manager instance.  This
    is intended to be used with :py:class:`Boto3ModelManager` classes that have
    a ``Policy`` attribute.
    """

    def not_in_use(
        self,
        policies: list["IAMPolicy"] | None = None,
    ) -> list["IAMPolicy"]:
        """
        Return a list of policies that are not currently in use.

        If ``policies`` is not specified, all policies will be filtered.

        Keyword Args:
            policies: A list of policies to filter.

        """
        if policies is None:
            policies = self.list().all()  # type: ignore[attr-defined]
        cast("list[IAMPolicy]", policies)
        return [policy for policy in policies if not policy.is_in_use]  # type: ignore[attr-defined, union-attr]

    def in_use(
        self,
        policies: list["IAMPolicy"] | None = None,
    ) -> list["IAMPolicy"]:
        """
        Return a list of policies that are in use.

        Keyword Args:
            policies: A list of policies to filter.

        """
        if policies is None:
            policies = self.list().all()  # type: ignore[attr-defined]
        cast("list[IAMPolicy]", policies)
        return [policy for policy in policies if policy.is_in_use]  # type: ignore[attr-defined, union-attr]



from __future__ import annotations

from datetime import datetime  # noqa: TC003
from typing import List

from pydantic import BaseModel, Field


class ECRPullThroughCacheAction(BaseModel):
    """
    Represents details about an ECR pull-through cache action.  Contains
    information about the pull-through cache operation, including status,
    repository details, and any failure information if the operation was
    unsuccessful.
    """

    #: The status of the cache synchronization operation (e.g., "SUCCESS", "FAILURE")
    sync_status: str = Field(..., alias="sync-status")
    #: The ECR repository prefix used for the pull-through cache configuration
    ecr_repository_prefix: str = Field(..., alias="ecr-repository-prefix")
    #: The name of the repository being cached
    repository_name: str = Field(..., alias="repository-name")
    #: The URL of the upstream registry from which images are being pulled
    upstream_registry_url: str = Field(..., alias="upstream-registry-url")
    #: The tag of the image being pulled through the cache, if applicable
    image_tag: str | None = Field(None, alias="image-tag")
    #: The SHA256 digest of the image being pulled through the cache, if applicable
    image_digest: str | None = Field(None, alias="image-digest")
    #: A code indicating the reason for failure, if the operation failed
    failure_code: str | None = Field(None, alias="failure-code")
    #: A human-readable description of why the operation failed, if applicable
    failure_reason: str | None = Field(None, alias="failure-reason")


class ECRPullThroughCacheActionEvent(BaseModel):
    """
    Represents a complete ECR Pull Through Cache Action event from EventBridge.
    Contains metadata about the event and the actual pull-through cache action
    details.
    """

    #: The detailed information about the ECR pull-through cache action
    detail: ECRPullThroughCacheAction
    #: The AWS account ID where the event occurred
    account: str
    #: The human-readable type of the event (e.g., "ECR Pull Through Cache Action")
    detail_type: str = Field(..., alias="detail-type")
    #: The unique identifier for this event instance
    id: str
    #: The AWS region where the event occurred
    region: str
    #: The AWS resources involved in this event, typically the repository ARN
    resources: List[str]
    #: The source of the event (typically "aws.ecr")
    source: str
    #: The timestamp when the event was generated, in UTC
    time: datetime
    #: The version of the event schema format
    version: str



from __future__ import annotations

from datetime import datetime  # noqa: TC003

from pydantic import BaseModel, Field, RootModel


class Details(BaseModel):
    """
    A details of a :py:class:`AttachmentDetails` object in the context of
    a ECS task state change event.
    """

    #: The name of the detail
    name: str | None = None
    #: The value of the detail
    value: str | None = None


class AttachmentDetails(BaseModel):
    """
    Details of the attachment, used to represent the details of an attachment in
    the context of an ECS task state change event.
    """

    #: The attachment ID
    id: str | None = None
    #: The type of the attachment
    type: str | None = None
    #: The status of the attachment
    status: str | None = None
    #: The details of the attachment
    details: list[Details] = Field(default_factory=list)


class AttributesDetails(BaseModel):
    """
    The details of an attribute in the context of an ECS task state change event.
    """

    #: The name of the attribute
    name: str | None = None
    #: The value of the attribute
    value: str | None = None


class NetworkInterfaceDetails(BaseModel):
    """
    The details of a network interface in the context of an ECS task state change event.
    """

    #: The private IPv4 address of the network interface
    privateIpv4Address: str | None = None
    # The private IPv6 address of the network interface
    ipv6Address: str | None = None
    #: The attachment ID of the network interface
    attachmentId: str | None = None


class NetworkBindingDetails(BaseModel):
    """
    Network binding details in the context of an ECS task state change event.
    """

    #: The host binding IP address
    bindIP: str | None = None
    #: The host binding protocol
    protocol: str | None = None
    #: The container port
    containerPort: float | None = None
    #: The host port
    hostPort: float | None = None


class Environment(RootModel[dict[str, str]]):
    pass


class ContainerDetails(BaseModel):
    """
    Details about a container in the task definition in the context of an ECS
    task state change event.
    """

    #: The full URL for the image used to run the container
    image: str | None = None
    #: The digest of the image used to run the container
    imageDigest: str | None = None
    #: The list of network interfaces attached to the container
    networkInterfaces: list[NetworkInterfaceDetails] = Field(default_factory=list)
    #: The list of network bindings for the container
    networkBindings: list[NetworkBindingDetails] = Field(default_factory=list)
    #: The minimum amount of memory required to the container
    memory: str | None = None
    #: The maximum amount of memory allowed to the container
    memoryReservation: str | None = None
    #: The ARN of the task that the container is part of
    taskArn: str
    #: The name of the container
    name: str
    #: The exit code of the container if it has stopped
    exitCode: float | None = None
    #: The maximum amount of CPU allowed to the container
    cpu: str | None = None
    #: The ARN of the container
    containerArn: str
    #: The last status of the container
    lastStatus: str
    #: The runtime ID of the container
    runtimeId: str | None = None
    #: The reason why the container has :py:attr:`lastStatus`
    reason: str | None = None
    #: A list of GPU IDs that the container is using
    gpuIds: list[str] | None = None


class OverridesItem(BaseModel):
    """
    Overrides for a single container in the task definition.
    """

    #: Overrides to environment variables
    environment: list[Environment] = Field(default_factory=list)
    #: Overrides to container memory
    memory: float | None = None
    #: The name of the container these overrides apply to
    name: str
    # Overrides to container CPU
    cpu: float | None = None
    #: Overrides to container command
    command: list[str] | None = None


class Overrides(BaseModel):
    """
    Overrides to the task definition in the context of an ECS task state change
    event.  These are specified in the ``RunTask`` API call.
    """

    #: Container overrides
    containerOverrides: list[OverridesItem] = Field(default_factory=list)


class ECSTaskStateChange(BaseModel):
    """
    The main details of the ECS task state change event.  This is the ``detail``
    field.  in the json of the event.
    """

    #: Container overrides
    overrides: Overrides
    #: If the task is stopped, the time it was stopped in UTC
    executionStoppedAt: datetime | None = None
    #: The task maximum amount of memory
    memory: str | None = None
    #: A list of attachments to the task
    attachments: list[AttachmentDetails] = Field(default_factory=list)
    #: A list of attributes for the task
    attributes: list[AttributesDetails] = Field(default_factory=list)
    #: The time the images for the started being pulled, in UTC
    pullStartedAt: datetime | None = None
    #: The ARN of the task
    taskArn: str
    #: The time the task was started, in UTC
    startedAt: datetime | None = None
    #: The time the task was created, in UTC
    createdAt: datetime
    #: The ARN of the cluster the task is running in
    clusterArn: str
    connectivity: str | None = None
    #: The platform version of the ECS infrastructure
    platformVersion: str | None = None
    #: The ARN of the container instance the task is running on, if any
    #: (Fargate tasks do not have a container instance)
    containerInstanceArn: str | None = None
    #: The launch type of the task
    #: (EC2 or Fargate)
    launchType: str | None = None
    #: ? not sure what this is
    group: str | None = None
    #: If the task was updated, the time it was updated in UTC
    updatedAt: datetime
    #: The AWS code describing why the task was stopped
    stopCode: str | None = None
    #: THe time the pulls of all the images for the task ended, in UTC
    pullStoppedAt: datetime | None = None
    #: The time connectivity was established, in UTC
    connectivityAt: datetime | None = None
    #: The ARN of the IAM role or user used that started the task
    startedBy: str | None = None
    #: The maximum amount of CPU allowed to the task
    cpu: str | None = None
    #: ? not sure what this is
    version: float
    #: The time we received the stop task request, in UTC
    stoppingAt: datetime | None = None
    #: The time the task was stopped, in UTC
    stoppedAt: datetime | None = None
    #: The ARN of the task definition used to start the task
    taskDefinitionArn: str
    #: The reason why the task was stopped
    stoppedReason: str | None = None
    #: The list of containers in the task
    containers: list[ContainerDetails] = Field(default_factory=list)
    #: The desired status of the task
    desiredStatus: str
    #: The current status of the task
    lastStatus: str
    #: The availability zone the task is running in
    availabilityZone: str | None = None


class ECSTaskStateChangeEvent(BaseModel):
    #: Where the meat of the event is
    detail: ECSTaskStateChange
    #: The human readable name of the event
    detail_type: str = Field(..., alias="detail-type")
    #: The resources the event applies to.  This will be a list of ARNs
    resources: list[str]
    #: The ID of the event
    id: str
    #: The source of the event, e.g. "aws.ecs"
    source: str
    #: The time the event was generated, in UTC
    time: datetime
    #: The region the event was generated in
    region: str
    #: The version of the schema for the event
    version: str
    #: The account the event was generated in
    account: str



from .aws_api_call_via_cloudtrail import ECRAWSAPICallViaCloudTrailEvent  # noqa: F401
from .ecr_image_action import ECRImageActionEvent  # noqa: F401
from .ecr_image_scan import ECRImageScanEvent  # noqa: F401
from .pull_through_cache_action import ECRPullThroughCacheActionEvent  # noqa: F401
from .referrer_action import ECRReferrerActionEvent  # noqa: F401
from .replication_action import ECRReplicationActionEvent  # noqa: F401
from .scan_resource_change import ECRScanResourceChangeEvent  # noqa: F401



# mypy: disable-error-code="attr-defined"
from functools import wraps
from typing import Callable

from botocraft.services.abstract import PrimaryBoto3ModelQuerySet

# ----------
# Decorators
# ----------


# Service


def kms_keys_only(
    func: Callable[..., "PrimaryBoto3ModelQuerySet"],
) -> Callable[..., "PrimaryBoto3ModelQuerySet"]:
    """
    Wraps :py:meth:`botocraft.services.kms.KMSKeyManager.list` to return a
    queryset of :py:class:`botocraft.services.kms.KMSKey` objects instead of
    only a list of key ids.
    """

    @wraps(func)
    def wrapper(self, *args, **kwargs) -> "PrimaryBoto3ModelQuerySet":
        _ids = func(self, *args, **kwargs)
        return PrimaryBoto3ModelQuerySet([self.get(KeyId=_id.KeyId) for _id in _ids])

    return wrapper



from functools import wraps
from typing import TYPE_CHECKING, Any, Callable, Generator, List, Optional, Union

from botocraft.services.abstract import PrimaryBoto3ModelQuerySet

if TYPE_CHECKING:
    from botocraft.eventbridge import AbstractEventFactory, EventBridgeEvent
    from botocraft.services.sqs import Message


# ----------
# Decorators
# ----------


def queue_list_urls_to_queues(
    func: Callable[..., List["str"]],
) -> Callable[..., "PrimaryBoto3ModelQuerySet"]:
    """
    Wraps a boto3 method that returns a list of SQS queue URLs to return a list
    of :py:class:`Queue` objects instead.
    """

    @wraps(func)
    def wrapper(*args, **kwargs) -> "PrimaryBoto3ModelQuerySet":
        self = args[0]
        urls = func(*args, **kwargs)
        names = [url.split("/")[-1] for url in urls]
        return PrimaryBoto3ModelQuerySet([self.get(QueueName=name) for name in names])

    return wrapper


def queue_recieve_messages_add_queue_url(
    func: Callable[..., List["Message"]],
) -> Callable[..., "PrimaryBoto3ModelQuerySet"]:
    """
    Wraps a boto3 method that receives messages from an SQS queue to return
    a queryset of :py:class:`~botocraft.services.sqs.Message` objects with the
    queue URL added.
    """

    @wraps(func)
    def wrapper(*args, **kwargs) -> "PrimaryBoto3ModelQuerySet":
        queue_url = args[1]
        messages = func(*args, **kwargs)

        if not messages:
            return PrimaryBoto3ModelQuerySet([])
        for message in messages:
            message.QueueUrl = queue_url
        return PrimaryBoto3ModelQuerySet(messages)  # type: ignore[arg-type]

    return wrapper


def queue_recieve_messages_add_event_factory(
    func: Callable[..., List["Message"]],
) -> Callable[..., "PrimaryBoto3ModelQuerySet"]:
    """
    Wraps a boto3 method that receives messages from an SQS queue to return
    a queryset of :py:class:`~botocraft.services.sqs.Message` objects with the
    EventFactoryClass added.  This is useful for converting the message body
    to an event object.
    """

    @wraps(func)
    def wrapper(*args, **kwargs) -> "PrimaryBoto3ModelQuerySet":
        from botocraft.eventbridge import EventFactory

        event_factory_class = kwargs.pop("EventFactoryClass", None)
        messages = func(*args, **kwargs)

        if not messages:
            return PrimaryBoto3ModelQuerySet([])
        if not event_factory_class:
            event_factory_class = EventFactory
        for message in messages:
            message.EventFactoryClass = event_factory_class
        return PrimaryBoto3ModelQuerySet(messages)  # type: ignore[arg-type]

    return wrapper


# -------------
# Mixin Classes
# -------------


class QueueManagerMixin:
    """
    A mixin class that extends :py:class:`~botocraft.services.sqs.QueueManager`
    to add the :py:meth:`get` method to retrieve a queue by name.   Queues are
    not first class objects in AWS SQS, so this is a convenience method to
    retrieve a queue by name and return our bespoke
    :py:class:`~botocraft.service.sqs.Queue` object.
    """

    def get(self, QueueName: str):  # noqa: N803
        """
        Get a queue by name.

        Args:
            QueueName: The name of the queue to retrieve.

        Raises:
            botocore.exceptions.ClientError: If the queue does not exist or if
              there is an error retrieving it.

        Returns:
            An object representing the queue, including its URL,
              attributes, and tags.

        """
        from botocraft.services import Queue, Tag

        sqs = self.client  # type: ignore[attr-defined]
        response = sqs.get_queue_url(QueueName=QueueName)
        queue_url = response["QueueUrl"]
        response = sqs.get_queue_attributes(
            QueueUrl=queue_url,
            AttributeNames=["All"],
        )
        attributes = response["Attributes"]
        tags = sqs.list_queue_tags(QueueUrl=queue_url)
        # Unfortunately the tags are returned as a dict with the key "Tags" and
        # the value being a dict of tags. We need to extract the tags from this
        # dict and convert them to a list of dicts, like TagsDictMixin expects
        if "Tags" not in tags:
            tags["Tags"] = []
        else:
            tags["Tags"] = [Tag(Key=k, Value=v) for k, v in tags["Tags"].items()]

        queue = Queue(
            QueueName=QueueName,
            QueueUrl=queue_url,
            Attributes=attributes if attributes else None,
            Tags=tags["Tags"],
        )
        queue.set_session(self.session)  # type: ignore[attr-defined]
        return queue


class QueueModelMixin:
    """
    A mixin class that extends :py:class:`~botocraft.services.sqs.Queue`
    to provide a generator that will yield all the messages in the queue
    eternally.  This is useful for a job that needs to listen continuously
    for messages on a queue.
    """

    def poll(
        self,
        EventFactoryClass: Optional["AbstractEventFactory"] = None,  # noqa: N803
    ) -> Generator["Message", None, None]:
        """
        Eternally poll for messages in the queue, and yield them as
        :py:class:`~botocraft.eventbridge.EventBridgeEvent` objects or dicts (if
        we can't identify the event) as they arrive.  This is useful for a job
        that needs to listen continuously for messages on a queue.

        Keyword Args:
            EventFactoryClass: The class to use to convert the message body
                to an event object.  If not provided, the default
                :py:class:`~botocraft.eventbridge.EventFactory` class will be used.

        Yields:
            A :py:class:`~botocraft.services.sqs.Message` from the queue

        """
        while True:
            messages = self.receive(  # type: ignore[attr-defined]
                MaxNumberOfMessages=10,
                WaitTimeSeconds=20,
                EventFactoryClass=EventFactoryClass,  # type: ignore[attr-defined]
            )
            if not messages:
                continue
            yield from messages


class MessageModelMixin:
    """
    A mixin class that extends :py:class:`~botocraft.services.sqs.Message`
    to provide a method to convert the message body to an event object.
    """

    @property
    def event(self) -> Union["EventBridgeEvent", dict[str, Any]]:
        """
        Convert the message body to an event object using the
        :py:class:`~botocraft.eventbridge.EventFactory` class.

        Returns:
            An event object or dict representing the message body.

        """
        if not hasattr(self, "EventFactoryClass"):
            msg = "EventFactoryClass is not set on the message."
            raise ValueError(msg)
        return self.EventFactoryClass().new(self.Body)  # type: ignore[attr-defined]



from __future__ import annotations

from datetime import datetime  # noqa: TC003
from typing import Any, Dict, List

from pydantic import BaseModel, Field


class RequestParametersItem(BaseModel):
    """Represents a placement constraint or other parameter for an ECS request."""

    #: The expression for the placement constraint
    expression: str
    #: The type of the placement constraint
    type: str


class RequestParametersItem1(BaseModel):
    """Represents a container definition in the ECS request parameters."""

    #: The name of the container
    containerName: str
    #: The exit code of the container, if it has exited
    exitCode: float | None = None
    #: The network bindings for the container
    networkBindings: List[Dict[str, Any]] | None = None
    #: The status of the container
    status: str


class ResponseElementsItemItem1(BaseModel):
    """Represents a container definition in the ECS response elements."""

    #: The Docker image used by the container
    image: str
    #: The network interfaces attached to the container
    networkInterfaces: List[Dict[str, Any]]
    #: The memory allocated to the container
    memory: str | None = None
    #: The ARN of the task this container belongs to
    taskArn: str
    #: The name of the container
    name: str
    #: The CPU units allocated to the container
    cpu: str
    #: The ARN of the container
    containerArn: str
    #: The current status of the container
    lastStatus: str


class AwsvpcConfiguration(BaseModel):
    """Represents the VPC configuration for an ECS task."""

    #: Whether to assign a public IP address to the task
    assignPublicIp: str | None = None
    #: The subnets in which to launch the task
    subnets: List[str] | None = None


class OverridesItem(BaseModel):
    """Represents overrides for a container in the ECS request."""

    #: The resource requirements for the container
    resourceRequirements: List[Dict[str, Any]] | None = None
    #: The environment variables for the container
    environment: List[Dict[str, Any]] | None = None
    #: The memory override for the container
    memory: float | None = None
    #: The name of the container to override
    name: str
    #: The CPU override for the container
    cpu: float | None = None
    #: The command override for the container
    command: List[str] | None = None


class SessionIssuer(BaseModel):
    """Represents the issuer of a session in the user identity context."""

    #: The account ID of the session issuer
    accountId: str
    #: The principal ID of the session issuer
    principalId: str
    #: The type of the session issuer
    type: str
    #: The ARN of the session issuer
    arn: str
    #: The username of the session issuer
    userName: str


class Attributes(BaseModel):
    """Represents session attributes, such as MFA status and creation date."""

    #: Whether MFA was used for authentication
    mfaAuthenticated: str
    #: When the session was created
    creationDate: datetime


class Overrides1Item(BaseModel):
    """Represents overrides for a container in the ECS response."""

    #: The resource requirements for the container
    resourceRequirements: List[Dict[str, Any]] | None = None
    #: The environment variables for the container
    environment: List[Dict[str, Any]] | None = None
    #: The memory override for the container
    memory: float | None = None
    #: The name of the container to override
    name: str
    #: The CPU override for the container
    cpu: float | None = None
    #: The command override for the container
    command: List[str] | None = None


class ResponseElementsItemItemItem(BaseModel):
    """Represents a detail item for an ECS response element attachment."""

    #: The name of the detail
    name: str
    #: The value of the detail
    value: str


class NetworkConfiguration(BaseModel):
    """Represents the network configuration for an ECS task."""

    #: The AWS VPC configuration for the task
    awsvpcConfiguration: AwsvpcConfiguration | None = None


class Overrides(BaseModel):
    """Represents container overrides for an ECS request."""

    #: The overrides for individual containers
    containerOverrides: List[OverridesItem] | None = None


class SessionContext(BaseModel):
    """Represents the session context for a user identity."""

    #: Web identity federation data, if applicable
    webIdFederationData: Dict[str, Any] | None = None
    #: Information about the entity that issued the session
    sessionIssuer: SessionIssuer
    #: Session attributes like MFA status
    attributes: Attributes


class Overrides1(BaseModel):
    """Represents container and inference accelerator overrides for an ECS response."""

    #: The overrides for individual containers
    containerOverrides: List[Overrides1Item]
    #: The overrides for inference accelerators
    inferenceAcceleratorOverrides: List[Dict[str, Any]]


class ResponseElementsItemItem(BaseModel):
    """Represents an attachment for an ECS response element."""

    #: The details of the attachment
    details: List[ResponseElementsItemItemItem]
    #: The ID of the attachment
    id: str
    #: The type of the attachment
    type: str
    #: The status of the attachment
    status: str


class RequestParameters(BaseModel):
    """Represents the parameters sent in an ECS API request via CloudTrail."""

    #: The network configuration for the task
    networkConfiguration: NetworkConfiguration | None = None
    #: Overrides for the task
    overrides: Overrides | None = None
    #: Placement constraints for the task
    placementConstraints: List[RequestParametersItem] | None = None
    #: The cluster on which to run the task
    cluster: str | None = None
    #: The reason for the request
    reason: str | None = None
    #: When task execution stopped, if applicable
    executionStoppedAt: str | None = None
    #: When image pulling stopped, if applicable
    pullStoppedAt: str | None = None
    #: The entity that started the task
    startedBy: str | None = None
    #: When image pulling started, if applicable
    pullStartedAt: str | None = None
    #: The number of tasks to start
    count: float | None = None
    #: The task ID or ARN
    task: str | None = None
    #: The container instance on which to run the task
    containerInstance: str | None = None
    #: The containers to include in the task
    containers: List[RequestParametersItem1] | None = None
    #: The task definition to use
    taskDefinition: str | None = None
    #: Whether to enable ECS managed tags
    enableECSManagedTags: bool | None = None
    #: The launch type for the task (EC2 or FARGATE)
    launchType: str | None = None
    #: The desired status of the task
    status: str | None = None


class UserIdentity(BaseModel):
    """Represents the identity of the user that initiated the ECS API call."""

    #: The context of the session
    sessionContext: SessionContext | None = None
    #: The access key ID used for the request
    accessKeyId: str
    #: The AWS account ID of the user
    accountId: str
    #: The principal ID of the user
    principalId: str
    #: The type of the identity
    type: str
    #: The ARN of the identity
    arn: str
    #: The AWS service that invoked the API call, if applicable
    invokedBy: str | None = None


class ResponseElementsItem(BaseModel):
    """Represents a single ECS task or resource returned in the ECS API response."""

    #: Task overrides
    overrides: Overrides1
    #: The task memory
    memory: str
    #: Attachments to the task
    attachments: List[ResponseElementsItemItem]
    #: The ARN of the role or user that submitted the request
    startedBy: str | None = None
    #: The ARN of the task
    taskArn: str
    #: The CPU of the task
    cpu: str
    #: The version number of the task
    version: float
    #: The tags associated with the task
    tags: List[Dict[str, Any]]
    #: When the task was created
    createdAt: str
    #: The ARN of the cluster in which the task is running
    clusterArn: str
    #: The ARN of the task definition for the task
    taskDefinitionArn: str
    #: The platform version (FARGATE only)
    platformVersion: str | None = None
    #: A list of container definitions
    containers: List[ResponseElementsItemItem1]
    #: The ARN of the container instance on which the task is running
    #: (Fargate tasks have no container instance)
    containerInstanceArn: str | None = None
    #: The last status of the task
    lastStatus: str
    #: The desired status of the task
    desiredStatus: str
    #: The ECS group for the task
    group: str
    #: The launch type of the task (e.g., EC2, FARGATE)
    launchType: str


class ResponseElements(BaseModel):
    """Represents the response elements returned by the ECS API call."""

    #: The endpoint URL
    endpoint: str | None = None
    #: Any failures that occurred during the API call
    failures: List[Dict[str, Any]] | None = None
    #: The telemetry endpoint URL
    telemetryEndpoint: str | None = None
    #: The tasks returned in the response
    tasks: List[ResponseElementsItem] | None = None
    #: The acknowledgment message, if any
    acknowledgment: str | None = None


class AWSAPICallViaCloudTrail(BaseModel):
    """
    Represents the detail of an AWS API call event as recorded by CloudTrail for ECS.
    """

    #: The API response
    responseElements: ResponseElements | None = None
    #: The API request
    requestParameters: RequestParameters | None = None
    #: The identity of the user that initiated the event
    userIdentity: UserIdentity
    #: The ID of the event
    eventID: str
    #: The name of the region in which the event occurred
    awsRegion: str
    #: The version of the event
    eventVersion: str
    #: The source IP of the user that initiated the event
    sourceIPAddress: str
    #: The source of the event
    eventSource: str
    #: The UserAgent of the request
    userAgent: str
    #: The event type
    eventType: str
    #: The request ID of the event in CloudTrail
    requestID: str
    #: The time the event was created
    eventTime: datetime
    #: The name of the event
    eventName: str


class ECSAWSAPICallViaCloudTrailEvent(BaseModel):
    """
    Represents a full ECS API call event as received from EventBridge, wrapping
    the CloudTrail detail.
    """

    #: Where the meat of the event is
    detail: AWSAPICallViaCloudTrail
    #: The human readable name of the event
    detail_type: str = Field(..., alias="detail-type")
    #: The resources the event applies to. This will be a list of ARNs
    resources: list[str]
    #: The ID of the event
    id: str
    #: The source of the event, e.g. "aws.ecs"
    source: str
    #: The time the event was generated, in UTC
    time: datetime
    #: The region the event was generated in
    region: str
    #: The version of the schema for the event
    version: str
    #: The account the event was generated in
    account: str
